import json
import logging
import re
from collections.abc import Sequence
from pathlib import Path
from typing import Any, Union
from urllib.parse import urljoin

import geopandas as gpd
import numpy as np
import pandas as pd
import requests
from pyproj import CRS
from requests import Response
from requests.auth import HTTPBasicAuth

from roseau.load_flow.exceptions import RoseauLoadFlowException, RoseauLoadFlowExceptionCode
from roseau.load_flow.io import network_from_dgs, network_from_dict, network_to_dict
from roseau.load_flow.models import (
    AbstractBranch,
    AbstractBus,
    AbstractLoad,
    AbstractTransformer,
    Element,
    FlexibleLoad,
    Ground,
    PotentialRef,
    PowerLoad,
    VoltageSource,
)
from roseau.load_flow.utils import ureg

logger = logging.getLogger(__name__)


class ElectricalNetwork:
    DEFAULT_PRECISION: float = 1e-6
    DEFAULT_MAX_ITERATIONS: int = 20
    DEFAULT_BASE_URL = "https://load-flow-api.roseautechnologies.com/"

    # Default classes to use
    branch_class = AbstractBranch
    load_class = AbstractLoad
    bus_class = AbstractBus
    ground_class = Ground
    pref_class = PotentialRef

    #
    # Methods to build an electrical network
    #
    def __init__(
        self,
        buses: Union[list[AbstractBus], dict[Any, AbstractBus]],
        branches: Union[list[AbstractBranch], dict[Any, AbstractBranch]],
        loads: Union[list[AbstractLoad], dict[Any, AbstractLoad]],
        special_elements: list[Element],
        **kwargs,
    ) -> None:
        """ElectricalNetwork constructor

        Args:
            buses:
                The buses of the network

            branches:
                The branches of the network

            loads:
                The loads of the network

            special_elements:
                The other elements (special, ground...)
        """
        if isinstance(buses, list):
            buses_dict = dict()
            for bus in buses:
                if bus.id in buses_dict:
                    msg = f"Duplicate id for a bus in this network: {bus.id!r}."
                    logger.error(msg)
                    raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.DUPLICATE_BUS_ID)
                buses_dict[bus.id] = bus
            buses = buses_dict
        if isinstance(branches, list):
            branches_dict = dict()
            for branch in branches:
                if branch.id in branches_dict:
                    msg = f"Duplicate id for a branch in this network: {branch.id!r}."
                    logger.error(msg)
                    raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.DUPLICATE_BRANCH_ID)
                branches_dict[branch.id] = branch
            branches = branches_dict
        if isinstance(loads, list):
            loads_dict = dict()
            for load in loads:
                if load.id in loads_dict:
                    msg = f"Duplicate id for a load in this network: {load.id!r}."
                    logger.error(msg)
                    raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.DUPLICATE_LOAD_ID)
                loads_dict[load.id] = load
            loads = loads_dict

        self.buses: dict[Any, AbstractBus] = buses
        self.branches: dict[Any, AbstractBranch] = branches
        self.loads: dict[Any, AbstractLoad] = loads
        self.special_elements: list[Element] = special_elements

        self._check_validity(constructed=False)
        self._create_network()
        self._valid = True
        self._results_info: dict[str, Any] = dict()

    @classmethod
    def from_element(cls, initial_bus: AbstractBus) -> "ElectricalNetwork":
        """ElectricalNetwork constructor. Construct the network from only one element and add the others automatically

        Args:
            initial_bus:
                Any bus of the network
        """
        buses = []
        branches = []
        loads = []
        specials = []
        elements = [initial_bus]
        visited_elements = []
        while elements:
            e = elements.pop(-1)
            visited_elements.append(e)
            if isinstance(e, AbstractBus):
                buses.append(e)
            elif isinstance(e, AbstractBranch):
                branches.append(e)
            elif isinstance(e, AbstractLoad):
                loads.append(e)
            else:
                specials.append(e)
            for connected_element in e.connected_elements:
                if connected_element not in visited_elements and connected_element not in elements:
                    elements.append(connected_element)
        return cls(buses=buses, branches=branches, loads=loads, special_elements=specials)

    #
    # Methods to access the data
    #
    @property
    def buses_frame(self) -> gpd.GeoDataFrame:
        """A property to get a geo dataframe of the buses.

        Returns:
            The geo dataframe of buses.
        """
        return gpd.GeoDataFrame(
            data=pd.DataFrame.from_records(
                data=[(bus_id, bus.n, bus.geometry) for bus_id, bus in self.buses.items()],
                columns=["id", "n", "geometry"],
                index="id",
            ),
            geometry="geometry",
            crs=CRS("EPSG:4326"),
        )

    @property
    def branches_frame(self) -> gpd.GeoDataFrame:
        """A property to get a geo dataframe of the branches.

        Returns:
            The geo dataframe of branches.
        """
        return gpd.GeoDataFrame(
            data=pd.DataFrame.from_records(
                data=[
                    (
                        branch_id,
                        branch.branch_type,
                        branch.n1,
                        branch.n2,
                        branch.connected_elements[0].id,
                        branch.connected_elements[1].id,
                        branch.geometry,
                    )
                    for branch_id, branch in self.branches.items()
                ],
                columns=["id", "branch_type", "n1", "n2", "bus1_id", "bus2_id", "geometry"],
                index="id",
            ),
            geometry="geometry",
            crs=CRS("EPSG:4326"),
        )

    @property
    def loads_frame(self) -> pd.DataFrame:
        """A property to get a dataframe of the loads.

        Returns:
            The dataframe of loads.
        """
        return pd.DataFrame.from_records(
            data=[(load_id, load.n, load.bus.id) for load_id, load in self.loads.items()],
            columns=["id", "n", "bus_id"],
            index="id",
        )

    #
    # Method to solve a load flow
    #
    def solve_load_flow(
        self,
        auth: Union[tuple[str, str], HTTPBasicAuth],
        base_url: str = DEFAULT_BASE_URL,
        precision: float = DEFAULT_PRECISION,
        max_iterations: int = DEFAULT_MAX_ITERATIONS,
    ) -> int:
        """Execute a newton algorithm for load flow calculation. In order to get the results of the load flow, please
        use the `results` method or call the elements directly.

        Args:
            auth:
                The login and password for the roseau load flow api.

            base_url:
                The base url to request the load flow solver.

            precision:
                Precision needed for the convergence

            max_iterations:
                The maximum number of allowed iterations

        Returns:
            The number of iterations taken
        """
        if not self._valid:
            self._check_validity(constructed=True)
            self._create_network()

        # Get the data
        network_data = self.to_dict()

        # Request the server
        params = {"max_iterations": max_iterations, "precision": precision}
        response = requests.post(
            url=urljoin(base_url, "solve/"),
            params=params,
            json=network_data,
            auth=auth,
            headers={"accept": "application/json"},
        )

        # Read the response
        # HTTP 4xx,5xx
        if not response.ok:
            self._parse_error(response=response)

        # HTTP 200
        result_dict: dict[str, Any] = response.json()
        info = result_dict["info"]
        if info["status"] != "success":
            msg = (
                f"The load flow did not converge after {info['iterations']} iterations. The norm of the residuals is "
                f"{info['finalError']:.5n}"
            )
            logger.error(msg=msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.NO_LOAD_FLOW_CONVERGENCE)

        logger.info(
            f"The load flow converged after {info['iterations']} iterations (final error={info['finalError']:.5n})."
        )

        # Dispatch the results
        self._dispatch_results(result_dict=result_dict)

        return info["iterations"]

    @staticmethod
    def _parse_error(response: Response):
        """Parse a response when its status is not "ok".

        Args:
            response:
                The response to parse
        """
        content_type = response.headers.get("content-type", None)
        if response.status_code == 401:
            msg = "Authentication failed."
            logger.error(msg=msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_REQUEST)
        else:
            msg = f"There is a problem in the request. Error code {response.status_code}."
            if content_type == "application/json":
                result_dict: dict[str, Any] = response.json()

                # If we have a valid Roseau Load Flow Exception, raise it
                if "msg" in result_dict and "code" in result_dict:
                    try:
                        code = RoseauLoadFlowExceptionCode.from_string(result_dict["code"])
                    except Exception:
                        msg += f" {result_dict['code']} - {result_dict['msg']}"
                    else:
                        msg = result_dict["msg"]
                        logger.error(msg)
                        raise RoseauLoadFlowException(msg=msg, code=code)
            else:
                msg += response.text

            # Otherwise, raise a generic "Bad request"
            logger.error(msg=msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_REQUEST)

    def _dispatch_results(self, result_dict: dict[str, Any]) -> None:
        """Dispatch the results to all the elements of the network.

        Args:
            result_dict:
                The results returned by the solver.
        """
        for bus_data in result_dict["buses"]:
            bus_id = bus_data["id"]
            self.buses[bus_id].potentials = self._dispatch_value(bus_data["potentials"], "v")
        for branch_data in result_dict["branches"]:
            branch_id = branch_data["id"]
            currents1 = self._dispatch_value(branch_data["currents1"], "i")
            currents2 = self._dispatch_value(branch_data["currents2"], "i")
            self.branches[branch_id].currents = [currents1, currents2]
        for load_data in result_dict["loads"]:
            load_id = load_data["id"]
            load = self.loads[load_id]
            if isinstance(load, FlexibleLoad):
                load.powers = self._dispatch_value(load_data["powers"], "s")
            currents = self._dispatch_value(load_data["currents"], "i")
            load.currents = currents

    @staticmethod
    def _dispatch_value(value: dict[str, tuple[float, float]], t: str) -> np.ndarray:
        """Dispatch the currents from a dictionary to a list.

        Args:
            value:
                The dictionary value to dispatch.

            t:
                The type of value ("i", "v" or "s").

        Returns:
            The complex final value.
        """
        if t + "n" in value:
            res = [
                value[t + "a"][0] + 1j * value[t + "a"][1],
                value[t + "b"][0] + 1j * value[t + "b"][1],
                value[t + "c"][0] + 1j * value[t + "c"][1],
                value[t + "n"][0] + 1j * value[t + "n"][1],
            ]
        else:
            res = [
                value[t + "a"][0] + 1j * value[t + "a"][1],
                value[t + "b"][0] + 1j * value[t + "b"][1],
                value[t + "c"][0] + 1j * value[t + "c"][1],
            ]
        return np.asarray(res)

    #
    # Getter for the load flow results
    #
    @ureg.wraps("V", (None, None), strict=False)
    def bus_potentials(self, id: Any) -> np.ndarray:
        """Compute the potential of a bus

        Args:
            id:
                The id of the bus

        Returns:
            The complex value of the bus potential
        """
        return self.buses[id].potentials

    @property
    def buses_potentials(self) -> pd.DataFrame:
        """Get the potentials of buses after a load flow has been solved.

        Returns:
            The data frame of the potentials of the buses of the electrical network.
        """
        phases = ["a", "b", "c", "n"]
        potentials_dict = {"bus_id": [], "phase": [], "potential": []}
        for bus_id, bus in self.buses.items():
            potentials: np.ndarray = bus.potentials
            for i in range(len(potentials)):
                potentials_dict["bus_id"].append(bus_id)
                potentials_dict["phase"].append(phases[i])
                potentials_dict["potential"].append(potentials[i].m_as("V"))
        potentials_df: pd.DataFrame = (
            pd.DataFrame.from_dict(potentials_dict, orient="columns")
            .astype({"phase": pd.CategoricalDtype(categories=phases, ordered=True), "potential": complex})
            .set_index(["bus_id", "phase"])
        )
        return potentials_df

    @ureg.wraps(("A", "A"), (None, None), strict=False)
    def branch_currents(self, id: Any) -> tuple[np.ndarray, np.ndarray]:
        """Compute the current of a branch

        Args:
            id:
                The name of the branch

        Returns:
            The complex value of the branch current
        """
        return self.branches[id].currents

    @property
    def branches_currents(self) -> pd.DataFrame:
        """Get the currents of the branches after a load flow has been solved.

        Returns:
            The data frame of the currents of the branches of the electrical network.
        """
        phases = ["a", "b", "c", "n"]
        currents_dict = {"branch_id": [], "phase": [], "current1": [], "current2": []}
        for branch_id, branch in self.branches.items():
            currents1, currents2 = branch.currents
            nb_currents1, nb_currents2 = len(currents1), len(currents2)
            for j in range(max(nb_currents1, nb_currents2)):
                currents_dict["branch_id"].append(branch_id)
                currents_dict["phase"].append(phases[j])
                try:
                    currents_dict["current1"].append(currents1[j].m_as("A"))
                except IndexError:
                    currents_dict["current1"].append(np.nan)
                try:
                    currents_dict["current2"].append(currents2[j].m_as("A"))
                except IndexError:
                    currents_dict["current2"].append(np.nan)
        currents_df = (
            pd.DataFrame.from_dict(currents_dict, orient="columns")
            .astype(
                {
                    "phase": pd.CategoricalDtype(categories=phases, ordered=True),
                    "current1": complex,
                    "current2": complex,
                }
            )
            .set_index(["branch_id", "phase"])
        )
        return currents_df

    @ureg.wraps("A", (None, None), strict=False)
    def load_currents(self, id: Any) -> np.ndarray:
        """Compute the currents of a load

        Args:
            id:
                The id of the load

        Returns:
            The complex value of the branch current
        """
        return self.loads[id].currents

    @property
    def loads_currents(self) -> pd.DataFrame:
        """Get the currents of the loads after a load flow has been solved.

        Returns:
            The data frame of the currents of the loads of the electrical network.
        """
        phases = ["a", "b", "c", "n"]
        loads_dict = {"load_id": [], "phase": [], "current": []}
        for load_id, load in self.loads.items():
            currents = load.currents
            for i in range(len(currents)):
                loads_dict["load_id"].append(load_id)
                loads_dict["phase"].append(phases[i])
                loads_dict["current"].append(currents[i].m_as("A"))
        currents_df: pd.DataFrame = (
            pd.DataFrame.from_dict(loads_dict, orient="columns")
            .astype({"phase": pd.CategoricalDtype(categories=phases, ordered=True), "current": complex})
            .set_index(["load_id", "phase"])
        )
        return currents_df

    @property
    def loads_powers(self) -> pd.DataFrame:
        """Get the powers of the loads after a load flow has been solved. Only for FlexibleLoads.

        Returns:
            The data frame of the powers of the loads of the electrical network.
        """
        phases = ["a", "b", "c", "n"]
        loads_dict = {"load_id": [], "phase": [], "power": []}
        for load_id, load in self.loads.items():
            if isinstance(load, FlexibleLoad):
                powers = load.powers
                for i in range(len(powers)):
                    loads_dict["load_id"].append(load_id)
                    loads_dict["phase"].append(phases[i])
                    loads_dict["power"].append(powers[i].m_as("VA"))
        powers_df: pd.DataFrame = (
            pd.DataFrame.from_dict(loads_dict, orient="columns")
            .astype({"phase": pd.CategoricalDtype(categories=phases, ordered=True), "power": complex})
            .set_index(["load_id", "phase"])
        )
        return powers_df

    #
    # Set the dynamic parameters.
    #
    def set_load_point(self, load_point: dict[Any, Sequence[complex]]) -> None:
        """Set a new load point to the network

        Args:
            load_point:
                The new load point
        """
        for load_id, value in load_point.items():
            load: AbstractLoad = self.loads[load_id]
            if isinstance(load, PowerLoad) or isinstance(load, FlexibleLoad):
                load.update_powers(value)
            else:
                msg = "Only power loads can be updated yet..."
                logger.error(msg)
                raise NotImplementedError(msg)

    def set_source_voltages(self, voltages: dict[Any, Sequence[complex]]) -> None:
        """Set new voltages for the voltage source

        Args:
            voltages:
                A dictionary voltage_source_id -> voltages to update
        """
        for bus_id, value in voltages.items():
            voltage_source = self.buses[bus_id]
            if not isinstance(voltage_source, VoltageSource):
                msg = "Only voltage sources can have their voltages updated."
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_ELEMENT_OBJECT)
            voltage_source.update_source_voltages(source_voltages=value)

    def add_element(self, element: Element) -> None:
        """Add an element to the network (the C++ electrical network and the tape will be recomputed).

        Args:
            element:
                The element to add.
        """
        if isinstance(element, AbstractBus):
            self.buses[element.id] = element
        elif isinstance(element, AbstractLoad):
            self.loads[element.id] = element
        elif isinstance(element, AbstractBranch):
            self.branches[element.id] = element
        else:
            msg = "Only lines, loads and buses can be added to the network."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_ELEMENT_OBJECT)
        self._valid = False

    def remove_element(self, id: Any) -> None:
        """Remove an element of the network (the C++ electrical network and the tape will be recomputed).

        Args:
            id:
                The id of the element to remove.
        """
        if id in self.buses:
            bus = self.buses.pop(id)
            bus.disconnect()
        elif id in self.loads:
            load = self.loads.pop(id)
            load.disconnect()
        elif id in self.branches:
            branch = self.branches.pop(id)
            branch.disconnect()
        else:
            msg = f"{id!r} is not a valid bus, branch or load id."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_ELEMENT_ID)
        self._valid = False

    def _create_network(self) -> None:
        """Create the Cython and C++ electrical network of all the passed elements"""
        self._valid = True

    def _check_validity(self, constructed: bool) -> None:
        """Check the validity of the network to avoid having a singular jacobian matrix

        Args:
            constructed:
                True if the network is already constructed and we have added an element, False otherwise
        """
        elements: list[Element] = list(self.buses.values())
        elements += list(self.branches.values())
        elements += list(self.loads.values())
        elements += self.special_elements

        for element in elements:
            for adj_element in element.connected_elements:
                if adj_element not in elements:
                    try:
                        adj_id = adj_element.id
                    except AttributeError:
                        adj_id = type(adj_element).__name__
                    try:
                        element_id = element.id
                    except AttributeError:
                        element_id = type(element).__name__
                    if constructed:
                        msg = (
                            f"The element {adj_id!r} is connected to {element_id!r} but is not in the "
                            f"ElectricalNetwork constructor."
                        )
                    else:
                        msg = (
                            f"The element {adj_id!r} is connected to {element_id!r} but has not been added to the "
                            f"network, you should add it with 'add_element'."
                        )
                    logger.error(msg)
                    raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.UNKNOWN_ELEMENT)

        found_source = False
        for element in elements:
            if isinstance(element, VoltageSource):
                found_source = True
        if not found_source:
            msg = "There is no voltage source provided in the network, you must provide at least one."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.NO_VOLTAGE_SOURCE)

        self._check_ref(elements)

    @staticmethod
    def _check_ref(elements) -> None:
        """Check the number of potential references to avoid having a singular jacobian matrix"""
        visited_elements = []
        for initial_element in elements:
            if initial_element in visited_elements or isinstance(initial_element, AbstractTransformer):
                continue
            visited_elements.append(initial_element)
            connected_component = []
            to_visit = [initial_element]
            while to_visit:
                element = to_visit.pop(-1)
                connected_component.append(element)
                for connected_element in element.connected_elements:
                    if connected_element not in visited_elements and not isinstance(
                        connected_element, AbstractTransformer
                    ):
                        to_visit.append(connected_element)
                        visited_elements.append(connected_element)

            potential_ref = 0
            for element in connected_component:
                if isinstance(element, PotentialRef):
                    potential_ref += 1

            if potential_ref == 0:
                msg = (
                    f"The connected component containing the element {initial_element.id!r} does not have a "
                    f"potential reference."
                )
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.NO_POTENTIAL_REFERENCE)
            elif potential_ref >= 2:
                msg = (
                    f"The connected component containing the element {initial_element.id!r} has {potential_ref} "
                    f"potential references, it should have only one."
                )
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.SEVERAL_POTENTIAL_REFERENCE)

    #
    # Json Mixin interface
    #
    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "ElectricalNetwork":
        """ElectricalNetwork constructor from dict.

        Args:
            data:
                The dictionary containing the network data.

        Returns:
            The constructed network.
        """
        buses_dict, branches_dict, loads_dict, special_elements = network_from_dict(data=data, en_class=cls)
        return cls(buses=buses_dict, branches=branches_dict, loads=loads_dict, special_elements=special_elements)

    @classmethod
    def from_json(cls, path: Union[str, Path]) -> "ElectricalNetwork":
        """ElectricalNetwork constructor from json.

        Args:
            path:
                The path to the network data.

        Returns:
            The constructed network.
        """
        if not isinstance(path, Path):
            path = Path(path)
        data = json.loads(path.read_text())
        return cls.from_dict(data=data)

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary of the current network data.

        Returns:
            The created dictionary.
        """
        return network_to_dict(self)

    def to_json(self, path: Union[str, Path]) -> None:
        """Save the current network to a json file.

        Args:
            path:
                The path for the network output.
        """
        res = self.to_dict()
        output = json.dumps(res, ensure_ascii=False, indent=4)
        output = re.sub(r"\[\s+(.*),\s+(.*)\s+]", r"[\1, \2]", output)
        if not isinstance(path, Path):
            path = Path(path)
        path.write_text(output)

    #
    # Output of results
    #
    def results_to_dict(self) -> dict[str, Any]:
        """Get the voltages and currents computed by the load flow and return them as dict.

        Returns:
            The dictionary of the voltages of the buses, and the dictionary of the current flowing through the branches.
        """
        phases = ["a", "b", "c", "n"]
        buses_results = list()
        for bus_id, bus in self.buses.items():
            potentials: np.ndarray = bus.potentials
            potentials_dict = dict()
            for i in range(len(potentials)):
                potentials_dict[f"v{phases[i]}"] = [potentials[i].real.magnitude, potentials[i].imag.magnitude]
            buses_results.append({"id": bus_id, "potentials": potentials_dict})

        branches_results = list()
        for branch_id, branch in self.branches.items():
            currents1: np.ndarray = branch.currents[0]
            currents2: np.ndarray = branch.currents[1]
            currents_dict1 = dict()
            currents_dict2 = dict()
            for i in range(len(currents1)):
                currents_dict1[f"i{phases[i]}"] = [currents1[i].real.magnitude, currents1[i].imag.magnitude]
            for i in range(len(currents2)):
                currents_dict2[f"i{phases[i]}"] = [currents2[i].real.magnitude, currents2[i].imag.magnitude]
            branches_results.append({"id": branch_id, "currents1": currents_dict1, "currents2": currents_dict2})

        loads_results = list()
        for load_id, load in self.loads.items():
            currents: np.ndarray = load.currents
            currents_dict = dict()
            for i in range(len(currents)):
                currents_dict[f"i{phases[i]}"] = [currents[i].real.magnitude, currents[i].imag.magnitude]
            if isinstance(load, FlexibleLoad):
                powers: np.ndarray = load.powers
                powers_dict = dict()
                for i in range(len(powers)):
                    powers_dict[f"s{phases[i]}"] = [powers[i].real.magnitude, powers[i].imag.magnitude]
                loads_results.append({"id": load_id, "powers": powers_dict, "currents": currents_dict})
            else:
                loads_results.append({"id": load_id, "currents": currents_dict})

        return {
            "info": self._results_info,
            "buses": buses_results,
            "branches": branches_results,
            "loads": loads_results,
        }

    def results_to_json(self, path: Union[str, Path]) -> Path:
        """Write a json containing the voltages and currents results.

        Args:
            path:
                The path to write the json.

        Returns:
            The path it has been writen in.
        """
        if isinstance(path, str):
            path = Path(path).expanduser().resolve()

        dict_results = self.results_to_dict()
        output = json.dumps(dict_results, indent=4)
        output = re.sub(r"\[\s+(.*),\s+(.*)\s+]", r"[\1, \2]", output)
        path.write_text(output)
        return path

    #
    # DGS interface
    #
    @classmethod
    def from_dgs(cls, path: Union[str, Path]) -> "ElectricalNetwork":
        """ElectricalNetwork constructor from json dgs file (PowerFactory).

        Args:
            path:
                The path to the network data.

        Returns:
            The constructed network.
        """
        buses_dict, branches_dict, loads_dict, special_elements = network_from_dgs(filename=path)
        return cls(buses=buses_dict, branches=branches_dict, loads=loads_dict, special_elements=special_elements)



import json
import logging
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd

from roseau.load_flow.exceptions import RoseauLoadFlowException, RoseauLoadFlowExceptionCode
from roseau.load_flow.models import (
    AbstractBranch,
    AbstractBus,
    AbstractLine,
    AbstractTransformer,
    Bus,
    Ground,
    LineCharacteristics,
    PotentialRef,
    PowerLoad,
    Switch,
    TransformerCharacteristics,
    VoltageSource,
)
from roseau.load_flow.models.core import Element
from roseau.load_flow.models.loads.loads import AbstractLoad
from roseau.load_flow.utils import LineModel, Q_

logger = logging.getLogger(__name__)


def network_from_dgs(  # noqa: C901
    filename: Union[str, Path]
) -> tuple[dict[str, AbstractBus], dict[str, AbstractBranch], dict[str, AbstractLoad], list[Element]]:
    """Create the electrical elements from a JSON file in DGS format to create an electrical network.

    Args:
        filename: name of the JSON file

    Returns:
        The buses, branches, loads and special elements to construct the electrical network.
    """
    # Read files
    (
        elm_xnet,
        elm_term,
        sta_cubic,
        elm_tr,
        typ_tr,
        elm_coup,
        elm_lne,
        typ_lne,
        elm_lod_lv,
        elm_lod_mv,
        elm_gen_stat,
        elm_pv_sys,
    ) = _read_dgs_json_file(filename=filename)

    # Ground and special elements
    ground = Ground()
    p_ref = PotentialRef(ground)

    # Buses
    buses = dict()
    for bus_id in elm_term.index:
        ph_tech = elm_term.at[bus_id, "phtech"]
        if ph_tech == 0:  # ABC
            n = 3
        elif ph_tech == 1:  # ABC-N
            n = 4
        else:
            msg = f"The Ph tech {ph_tech!r} for bus {bus_id!r} can not be handle."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.DGS_BAD_PHASE_TECHNOLOGY)
        buses[bus_id] = Bus(id=bus_id, n=n)

    # Sources
    for bus_id in elm_xnet.index:
        id_sta_cubic_source = elm_xnet.at[bus_id, "bus1"]  # id of the cubicle connecting the source and its bus
        source_bus = sta_cubic.at[id_sta_cubic_source, "cterm"]  # id of the bus to which the source is connected
        un = elm_term.at[source_bus, "uknom"] / np.sqrt(3) * 1e3  # phase-to-neutral voltage (V)
        tap = elm_xnet.at[bus_id, "usetp"]  # tap voltage (p.u.)
        # the voltage source "erases" the buse to which it is connected
        voltages = [un * tap, un * np.exp(-np.pi * 2 / 3 * 1j) * tap, un * np.exp(np.pi * 2 / 3 * 1j) * tap]
        buses[source_bus] = VoltageSource(
            id=source_bus,
            n=4,
            ground=ground,
            source_voltages=voltages,
        )

    # LV loads
    loads = dict()
    if elm_lod_lv is not None:
        _generate_loads(elm_lod_lv, loads, buses, sta_cubic, 1e3, production=False)

    # LV Production loads
    if elm_pv_sys is not None:
        _generate_loads(elm_pv_sys, loads, buses, sta_cubic, 1e3, production=True)
    if elm_gen_stat is not None:
        _generate_loads(elm_gen_stat, loads, buses, sta_cubic, 1e3, production=True)

    # MV loads
    if elm_lod_mv is not None:
        _generate_loads(elm_lod_mv, loads, buses, sta_cubic, 1e6, production=False)

    # Lines
    branches = dict()
    if elm_lne is not None:

        line_types = dict()
        for type_id in typ_lne.index:
            n = typ_lne.at[type_id, "nlnph"] + typ_lne.at[type_id, "nneutral"]
            if n == 4:
                line_model = LineModel.SYM_NEUTRAL
            elif n == 3:
                line_model = LineModel.SYM
            else:
                msg = f"The number of phases ({n}) of line type {type_id!r} can not be handled, it should be 3 or 4."
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.DGS_BAD_PHASE_NUMBER)

            line_types[type_id] = LineCharacteristics.from_sym(
                type_name=type_id,
                r0=typ_lne.at[type_id, "rline0"],
                model=line_model,
                r1=typ_lne.at[type_id, "rline"],
                x0=typ_lne.at[type_id, "xline0"],
                x1=typ_lne.at[type_id, "xline"],
                g0=Q_(typ_lne.at[type_id, "gline0"], "uS/km"),
                g1=Q_(typ_lne.at[type_id, "gline"], "uS/km"),
                b0=Q_(typ_lne.at[type_id, "bline0"], "uS/km"),
                b1=Q_(typ_lne.at[type_id, "bline"], "uS/km"),
                rn=typ_lne.at[type_id, "rnline"],
                xn=typ_lne.at[type_id, "xnline"],
                xpn=typ_lne.at[type_id, "xpnline"],
                bn=Q_(typ_lne.at[type_id, "bnline"], "uS/km"),
                bpn=Q_(typ_lne.at[type_id, "bpnline"], "uS/km"),
            )

        for line_id in elm_lne.index:
            type_id = elm_lne.at[line_id, "typ_id"]  # id of the line type

            branches[line_id] = AbstractLine.from_dict(
                id=line_id,
                bus1=buses[sta_cubic.at[elm_lne.at[line_id, "bus1"], "cterm"]],
                bus2=buses[sta_cubic.at[elm_lne.at[line_id, "bus2"], "cterm"]],
                length=elm_lne.at[line_id, "dline"],
                line_types=line_types,
                type_name=type_id,
                ground=ground,
            )

    # Transformers
    if elm_tr is not None:
        # Transformers type
        transformers_data = dict()
        transformers_tap = dict()
        for idx in typ_tr.index:
            # Extract data
            name = typ_tr.at[idx, "loc_name"]
            sn = typ_tr.at[idx, "strn"] * 1e6  # The nominal voltages of the transformer (MVA -> VA)
            uhv = typ_tr.at[idx, "utrn_h"] * 1e3  # Phase-to-phase nominal voltages of the high voltages side (kV -> V)
            ulv = typ_tr.at[idx, "utrn_l"] * 1e3  # Phase-to-phase nominal voltages of the low voltages side (kV -> V)
            i0 = typ_tr.at[idx, "curmg"] / 3 / 100  # Current during off-load test (%)
            p0 = typ_tr.at[idx, "pfe"] * 1e3 / 3  # Losses during off-load test (kW -> W)
            psc = typ_tr.at[idx, "pcutr"] * 1e3  # Losses during short circuit test (kW -> W)
            vsc = typ_tr.at[idx, "uktr"] / 100  # Voltages on LV side during short circuit test (%)
            type_name = "{}{}{}".format(  # Winding of the transformer
                typ_tr.at[idx, "tr2cn_h"], typ_tr.at[idx, "tr2cn_l"], typ_tr.at[idx, "nt2ag"]
            )

            # Generate transformer parameters
            transformers_data[idx] = TransformerCharacteristics(name, type_name, uhv, ulv, sn, p0, i0, psc, vsc)
            transformers_tap[idx] = typ_tr.at[idx, "dutap"]

        # Create transformers
        for idx in elm_tr.index:
            type_id = elm_tr.at[idx, "typ_id"]  # id of the line type
            tap = 1.0 + elm_tr.at[idx, "nntap"] * transformers_tap[type_id] / 100
            branches[idx] = AbstractTransformer.from_dict(
                id=idx,
                transformer_types=transformers_data,
                type_name=type_id,
                bus1=buses[sta_cubic.at[elm_tr.at[idx, "bushv"], "cterm"]],
                bus2=buses[sta_cubic.at[elm_tr.at[idx, "buslv"], "cterm"]],
                tap=tap,
            )
            ground.connect(bus=buses[sta_cubic.at[elm_tr.at[idx, "buslv"], "cterm"]])

    # Create switches
    if elm_coup is not None:
        for switch_id in elm_coup.index:
            n = elm_coup.at[switch_id, "nphase"] + elm_coup.at[switch_id, "nneutral"]
            branches[switch_id] = Switch(
                id=switch_id,
                n=n,
                bus1=buses[sta_cubic.at[elm_coup.at[switch_id, "bus1"], "cterm"]],
                bus2=buses[sta_cubic.at[elm_coup.at[switch_id, "bus2"], "cterm"]],
            )

    return buses, branches, loads, [p_ref, ground]


def _read_dgs_json_file(filename: Union[str, Path]):
    """Read a JSON file in DGS format.

    Args:
        filename: name of the JSON file

    Returns:
        elm_xnet: dataframe of external sources
        elm_term: dataframe of terminals (i.e. buses)
        sta_cubic: dataframe of cubicles
        elm_tr: dataframe of transformers
        typ_tr: dataframe of types of transformer
        elm_coup: dataframe of switches
        elm_lne: dataframe of electrical line
        typ_lne: dataframe of types of line
        elm_lod_lv: dataframe of LV loads
        elm_lod_mv: dataframe of MV loads
        elm_gen_stat: dataframe of generators
    """
    # Create dataframe from JSON file
    with open(filename, encoding="ISO-8859-10") as f:
        data = json.load(f)

    # External sources
    elm_xnet = pd.DataFrame(columns=data["ElmXnet"]["Attributes"], data=data["ElmXnet"]["Values"]).set_index("FID")

    # Terminals (buses)
    elm_term = pd.DataFrame(columns=data["ElmTerm"]["Attributes"], data=data["ElmTerm"]["Values"]).set_index("FID")

    # Cubicles
    sta_cubic = pd.DataFrame(columns=data["StaCubic"]["Attributes"], data=data["StaCubic"]["Values"]).set_index("FID")

    # Transformers
    if "ElmTr2" in data.keys():
        elm_tr = pd.DataFrame(columns=data["ElmTr2"]["Attributes"], data=data["ElmTr2"]["Values"]).set_index("FID")
    else:
        elm_tr = None

    # Transformer types
    if "TypTr2" in data.keys():
        typ_tr = pd.DataFrame(columns=data["TypTr2"]["Attributes"], data=data["TypTr2"]["Values"]).set_index("FID")
    else:
        typ_tr = None

    # Switch
    if "ElmCoup" in data.keys():
        elm_coup = pd.DataFrame(columns=data["ElmCoup"]["Attributes"], data=data["ElmCoup"]["Values"]).set_index("FID")
    else:
        elm_coup = None

    # Lines
    if "ElmLne" in data.keys():
        elm_lne = pd.DataFrame(columns=data["ElmLne"]["Attributes"], data=data["ElmLne"]["Values"]).set_index("FID")
    else:
        elm_lne = None

    # Line types
    if "TypLne" in data.keys():
        typ_lne = pd.DataFrame(columns=data["TypLne"]["Attributes"], data=data["TypLne"]["Values"]).set_index("FID")
    else:
        typ_lne = None

    # LV loads
    if "ElmLodLV" in data.keys():
        elm_lod_lv = pd.DataFrame(columns=data["ElmLodLV"]["Attributes"], data=data["ElmLodLV"]["Values"]).set_index(
            "FID"
        )
    else:
        elm_lod_lv = None

    # MV loads
    if "ElmLodmv" in data.keys():
        elm_lod_mv = pd.DataFrame(columns=data["ElmLodmv"]["Attributes"], data=data["ElmLodmv"]["Values"]).set_index(
            "FID"
        )
    else:
        elm_lod_mv = None

    # Generators
    if "ElmGenStat" in data.keys():
        elm_gen_stat = pd.DataFrame(
            columns=data["ElmGenStat"]["Attributes"], data=data["ElmGenStat"]["Values"]
        ).set_index("FID")
    else:
        elm_gen_stat = None

    # LV generators
    # Generators
    if "ElmPvsys" in data.keys():
        elm_pv_sys = pd.DataFrame(columns=data["ElmPvsys"]["Attributes"], data=data["ElmPvsys"]["Values"]).set_index(
            "FID"
        )
    else:
        elm_pv_sys = None

    return (
        elm_xnet,
        elm_term,
        sta_cubic,
        elm_tr,
        typ_tr,
        elm_coup,
        elm_lne,
        typ_lne,
        elm_lod_lv,
        elm_lod_mv,
        elm_gen_stat,
        elm_pv_sys,
    )


def _generate_loads(
    elm_lod: pd.DataFrame,
    loads: dict[str, AbstractLoad],
    buses: dict[str, AbstractBus],
    sta_cubic: pd.DataFrame,
    factor: float,
    production: bool,
):
    """Generate the loads of a given dataframe.

    Args:
        elm_lod:
            The dataframe of loads.

        loads:
            The dictionary the loads will be added to.

        buses:
            The dataframe of buses.

        sta_cubic:
            The dataframe of cubicles.

        factor:
            The factor to multiply the load power (ex: 1e3 for kVA -> VA)

        production:
             True for production loads, False otherwise.
    """
    for load_id in elm_lod.index:
        sta_cubic_id = elm_lod.at[load_id, "bus1"]  # id of the cubicle connecting the load and its bus
        bus_id = sta_cubic.at[sta_cubic_id, "cterm"]  # id of the bus to which the load is connected

        if production:
            s_phase = _compute_production_load_power(elm_lod, load_id, "") * factor
            sa = sb = sc = 0
        else:
            s_phase = _compute_load_power(elm_lod, load_id, "") * factor
            sa = _compute_load_power(elm_lod, load_id, "r") * factor
            sb = _compute_load_power(elm_lod, load_id, "s") * factor
            sc = _compute_load_power(elm_lod, load_id, "t") * factor

        if sa == 0 and sb == 0 and sc == 0:  # Balanced
            loads[load_id] = PowerLoad(id=load_id, n=4, bus=buses[bus_id], s=[s_phase / 3, s_phase / 3, s_phase / 3])
        else:  # Unbalanced
            loads[load_id] = PowerLoad(id=load_id, n=4, bus=buses[bus_id], s=[sa, sb, sc])


def _compute_load_power(elm_lod: pd.DataFrame, load_id: str, suffix: str) -> complex:
    """Compute a load power in PWF format.

    Args:
        elm_lod:
            The dataframe of loads.

        load_id:
            The load id.

        suffix:
            The phase of the load (empty for balanced loads, or r, s, t for phases a, b, c)

    Returns:
        The apparent power.
    """
    p = elm_lod.at[load_id, "plini" + suffix]
    q = np.sqrt(elm_lod.at[load_id, "slini" + suffix] ** 2 - elm_lod.at[load_id, "plini" + suffix] ** 2)
    q *= np.sign(p) * (-1 if elm_lod.at[load_id, "pf_recap" + suffix] else 1)
    return p + 1j * q


def _compute_production_load_power(elm_lod: pd.DataFrame, load_id: str, suffix: str) -> complex:
    """Compute a production load power in PWF format.

    Args:
        elm_lod:
            The dataframe of loads.

        load_id:
            The load id.

        suffix:
            The phase of the load (empty for balanced loads, or r, s, t for phases a, b, c)

    Returns:
        The apparent power.
    """
    p = elm_lod.at[load_id, "pgini" + suffix]
    q = elm_lod.at[load_id, "qgini" + suffix]
    q *= np.sign(p) * (-1 if elm_lod.at[load_id, "pf_recap" + suffix] else 1)
    return -(p + 1j * q)



__author__ = "Benoît Vinot, Victor Gouin, Florent Cadoux, Sébastien Vallet"
__copyright__ = "Roseau Technologies 2018-2022"
__credits__ = "Roseau Technologies"
__license__ = "Proprietary"
__maintainer__ = "Sébastien Vallet"
__email__ = "sebastien.vallet@roseautechnologies.com"
__status__ = "In development"
__url__ = "https://github.com/RoseauTechnologies/Roseau_Load_Flow/"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['load_flow',
 'load_flow.io',
 'load_flow.models',
 'load_flow.models.lines',
 'load_flow.models.loads',
 'load_flow.models.transformers',
 'load_flow.network',
 'load_flow.utils']

package_data = \
{'': ['*']}

install_requires = \
['Shapely>=1.8.0,<2.0.0',
 'geopandas>=0.10.2',
 'numpy>=1.21.5,<2.0.0',
 'pandas>=1.4.0,<2.0.0',
 'pint>=0.19.2',
 'pygeos>=0.12.0',
 'regex>=2022.1.18',
 'requests>=2.28.1,<3.0.0',
 'rich>=11.0.0']

setup_kwargs = {
    'name': 'roseau-load-flow',
    'version': '0.2.1',
    'description': 'Three-phase load flow solver',
    'long_description': '![CI](https://github.com/RoseauTechnologies/Roseau_Load_Flow/workflows/CI/badge.svg)\n\n# Roseau Load Flow #\n\n## Installation ##\n\nThe simplest way is to download the docker container attached to this repository and to start it. I will start a\nJupyterlab session with the package installed.\n\n## Usage ##\n\nThere are 2 main ways to execute a load flow with thunders:\n\n### From files ###\n\nBy giving path to the needed files:\n\n```python\nfrom roseau.load_flow import ElectricalNetwork\n\nen = ElectricalNetwork.from_dgs(path=path)  # DGS\n\nen = ElectricalNetwork.from_json(path=path)  # Json\n\nen.solve_load_flow(auth=("username", "password"))\n```\n\n### From code ###\n\nBy describing the network and its components, here is a simple example:\n\n```python\nfrom roseau.load_flow import Ground, VoltageSource, Bus, PowerLoad, PotentialRef, SimplifiedLine, ElectricalNetwork, LineCharacteristics\nimport numpy as np\n\nground = Ground()\nvn = 400 / np.sqrt(3)\nvoltages = [vn, vn * np.exp(-2 / 3 * np.pi * 1j), vn * np.exp(2 / 3 * np.pi * 1j)]\nvs = VoltageSource(\n    id="source",\n    n=4,\n    ground=ground,\n    source_voltages=voltages,\n)\nload_bus = Bus(id="load bus", n=4)\nload = PowerLoad(id="power load", n=4, bus=load_bus, s=[100 + 0j, 100 + 0j, 100 + 0j])\nline_characteristics = LineCharacteristics(type_name="test", z_line=np.eye(4, dtype=complex))\nline = SimplifiedLine(\n    id="line",\n    n=4,\n    bus1=vs,\n    bus2=load_bus,\n    line_characteristics=line_characteristics,\n    length=10  # km\n)\np_ref = PotentialRef(element=ground)\n\nen = ElectricalNetwork(buses=[vs, load_bus], branches=[line], loads=[load], special_elements=[p_ref, ground])\n# or\n# en = ElectricalNetwork.from_element(vs)\n\nen.solve_load_flow(auth=("username", "password"))\n```\n\n<!-- Local Variables: -->\n<!-- mode: gfm -->\n<!-- coding: utf-8-unix -->\n<!-- ispell-local-dictionary: "british" -->\n<!-- End: -->\n',
    'author': 'Sébastien Vallet',
    'author_email': 'sebastien.vallet@roseautechnologies.com',
    'maintainer': 'Sébastien Vallet',
    'maintainer_email': 'sebastien.vallet@roseautechnologies.com',
    'url': 'https://github.com/RoseauTechnologies/Roseau_Load_Flow/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



import logging
from typing import Any, TYPE_CHECKING

from roseau.load_flow.exceptions import RoseauLoadFlowException, RoseauLoadFlowExceptionCode
from roseau.load_flow.models import (
    AbstractBranch,
    AbstractBus,
    AbstractLine,
    AbstractLoad,
    AbstractTransformer,
    Element,
    LineCharacteristics,
    TransformerCharacteristics,
)

if TYPE_CHECKING:
    from roseau.load_flow.network import ElectricalNetwork

logger = logging.getLogger(__name__)


def network_from_dict(
    data: dict[str, Any], en_class: type["ElectricalNetwork"]
) -> tuple[dict[str, AbstractBus], dict[str, AbstractBranch], dict[str, AbstractLoad], list[Element]]:
    """Create the electrical elements from a dictionary to create an electrical network.

    Args:
        data:
            The dictionary containing the network data.

        en_class:
            The ElectricalNetwork class to create

    Returns:
        The buses, branches, loads and special elements to construct the electrical network.
    """
    line_types = dict()
    for line_data in data["line_types"]:
        type_name = line_data["name"]
        line_types[type_name] = LineCharacteristics.from_dict(line_data)

    transformer_types = dict()
    for transformer_data in data["transformer_types"]:
        type_name = transformer_data["name"]
        transformer_types[type_name] = TransformerCharacteristics.from_dict(transformer_data)

    ground = en_class.ground_class()
    special_elements = [ground, en_class.pref_class(element=ground)]
    buses_dict = dict()
    loads_dict = dict()
    for bus_data in data["buses"]:
        buses_dict[bus_data["id"]] = en_class.bus_class.from_dict(bus_data, ground)
        for load_data in bus_data["loads"]:
            loads_dict[load_data["id"]] = en_class.load_class.from_dict(load_data, buses_dict[bus_data["id"]])

    branches_dict = dict()
    for branch_data in data["branches"]:
        bus1 = buses_dict[branch_data["bus1"]]
        bus2 = buses_dict[branch_data["bus2"]]
        branches_dict[branch_data["id"]] = en_class.branch_class.from_dict(
            branch_data,
            bus1,
            bus2,
            ground,
            line_types,
            transformer_types,
        )
        if isinstance(branches_dict[branch_data["id"]], AbstractTransformer):
            if bus2.n == 4:
                ground.connect(bus2)
            else:
                special_elements.append(en_class.pref_class(element=bus2))

    return buses_dict, branches_dict, loads_dict, special_elements


def network_to_dict(en: "ElectricalNetwork") -> dict[str, Any]:
    """Return a dictionary of the current network data.

    Args:
        en:
            The electrical network.

    Returns:
        The created dictionary.
    """
    # Export the buses and the loads
    buses = list()
    for bus in en.buses.values():
        bus_dict = bus.to_dict()
        for load in bus.connected_elements:
            if isinstance(load, AbstractLoad):
                bus_dict["loads"].append(load.to_dict())
        buses.append(bus_dict)

    # Export the branches with their characteristics
    branches = list()
    line_characteristics_dict = dict()
    transformer_characteristics_dict = dict()
    for branch in en.branches.values():
        branches.append(branch.to_dict())
        if isinstance(branch, AbstractLine):
            type_name = branch.line_characteristics.type_name
            if (
                type_name in line_characteristics_dict
                and branch.line_characteristics != line_characteristics_dict[type_name]
            ):
                msg = f"There are line characteristics duplicates: {type_name}"
                logger.error(msg)
                raise RoseauLoadFlowException(
                    msg=msg, code=RoseauLoadFlowExceptionCode.JSON_LINE_CHARACTERISTICS_DUPLICATES
                )
            line_characteristics_dict[branch.line_characteristics.type_name] = branch.line_characteristics
        elif isinstance(branch, AbstractTransformer):
            type_name = branch.transformer_characteristics.type_name
            if (
                type_name in transformer_characteristics_dict
                and branch.transformer_characteristics != transformer_characteristics_dict[type_name]
            ):
                msg = f"There are transformer characteristics duplicates: {type_name}"
                logger.error(msg)
                raise RoseauLoadFlowException(
                    msg=msg, code=RoseauLoadFlowExceptionCode.JSON_TRANSFORMER_CHARACTERISTICS_DUPLICATES
                )
            transformer_characteristics_dict[type_name] = branch.transformer_characteristics

    # Line characteristics
    line_characteristics = list()
    for lc in line_characteristics_dict.values():
        line_characteristics.append(lc.to_dict())
    line_characteristics.sort(key=lambda x: x["name"])  # Always keep the same order

    # Transformer characteristics
    transformer_characteristics = list()
    for tc in transformer_characteristics_dict.values():
        transformer_characteristics.append(tc.to_dict())
    transformer_characteristics.sort(key=lambda x: x["name"])  # Always keep the same order

    return {
        "buses": buses,
        "branches": branches,
        "line_types": line_characteristics,
        "transformer_types": transformer_characteristics,
    }



from roseau.load_flow.models.transformers.transformers import (
    AbstractTransformer,
    DeltaDeltaTransformer,
    DeltaWyeTransformer,
    DeltaZigzagTransformer,
    TransformerCharacteristics,
    WyeDeltaTransformer,
    WyeWyeTransformer,
    WyeZigzagTransformer,
)

__all__ = [
    "AbstractTransformer",
    "WyeWyeTransformer",
    "DeltaWyeTransformer",
    "DeltaDeltaTransformer",
    "WyeDeltaTransformer",
    "WyeZigzagTransformer",
    "DeltaZigzagTransformer",
    "TransformerCharacteristics",
]



from roseau.load_flow.utils.constants import (
    DELTA_P,
    EPSILON_0,
    EPSILON_R,
    F,
    LV_MV_LIMIT,
    MU_0,
    MU_R,
    OMEGA,
    PI,
    RHO,
    TAN_D,
)
from roseau.load_flow.utils.types import BranchType, ConductorType, IsolationType, LineModel, LineType, TransformerType
from roseau.load_flow.utils.units import Q_, ureg

__all__ = [
    # Constants
    "PI",
    "MU_0",
    "EPSILON_0",
    "F",
    "OMEGA",
    "RHO",
    "MU_R",
    "DELTA_P",
    "TAN_D",
    "EPSILON_R",
    "LV_MV_LIMIT",
    # Types
    "LineType",
    "LineModel",
    "ConductorType",
    "IsolationType",
    "BranchType",
    "TransformerType",
    # Units
    "Q_",
    "ureg",
]



import logging
from typing import Any, Optional

from pint import Quantity
from shapely.geometry import LineString, Point
from shapely.geometry.base import BaseGeometry

from roseau.load_flow.exceptions import RoseauLoadFlowException, RoseauLoadFlowExceptionCode
from roseau.load_flow.models.buses import AbstractBus, VoltageSource
from roseau.load_flow.models.core import AbstractBranch, Ground
from roseau.load_flow.models.lines.line_characteristics import LineCharacteristics
from roseau.load_flow.utils.types import BranchType
from roseau.load_flow.utils.units import ureg

logger = logging.getLogger(__name__)


class Switch(AbstractBranch):
    """A general purpose switch branch."""

    branch_type = BranchType.SWITCH

    def __init__(
        self, id: Any, n: int, bus1: AbstractBus, bus2: AbstractBus, geometry: Optional[Point] = None, **kwargs
    ) -> None:
        """Switch constructor.

        Args:
            id:
                The id of the branch.

            n:
                The number of ports of the extremity buses.

            bus1:
                Bus to connect to the switch.

            bus2:
                Bus to connect to the switch.

            geometry:
                The geometry of the switch.
        """
        if geometry is not None and not isinstance(geometry, Point):
            msg = f"The geometry for a {type(self)} must be a point: {geometry.geom_type} provided."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_GEOMETRY_TYPE)
        super().__init__(id=id, n1=n, n2=n, bus1=bus1, bus2=bus2, geometry=geometry, **kwargs)
        self._check_elements()
        self._check_loop()

    def to_dict(self) -> dict[str, Any]:
        res = super().to_dict()
        res["type"] = "switch"
        return res

    def _check_loop(self):
        """Check that there are no switch loop, raise an exception if it is the case"""
        visited_1 = set()
        elements = [self.connected_elements[0]]
        while elements:
            element = elements.pop(-1)
            visited_1.add(element)
            for e in element.connected_elements:
                if e not in visited_1 and (isinstance(e, AbstractBus) or isinstance(e, Switch)) and e != self:
                    elements.append(e)
        visited_2 = set()
        elements = [self.connected_elements[1]]
        while elements:
            element = elements.pop(-1)
            visited_2.add(element)
            for e in element.connected_elements:
                if e not in visited_2 and (isinstance(e, AbstractBus) or isinstance(e, Switch)) and e != self:
                    elements.append(e)
        if visited_1.intersection(visited_2):
            msg = f"There is a loop of switch involving the switch {self.id!r}. It is not allowed."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.SWITCHES_LOOP)

    def _check_elements(self):
        """Check that we can connect both elements."""
        element1 = self.connected_elements[0]
        element2 = self.connected_elements[1]
        if isinstance(element1, VoltageSource) and isinstance(element2, VoltageSource):
            msg = (
                f"The voltage sources {element1.id!r} and {element2.id!r} are "
                f"connected with the switch {self.id!r}. It is not allowed."
            )
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_VOLTAGES_SOURCES_CONNECTION)


class AbstractLine(AbstractBranch):
    """An abstract class for all lines of this package."""

    branch_type = BranchType.LINE
    _simplified_line_class: Optional[type["SimplifiedLine"]] = None
    _shunt_line_class: Optional[type["ShuntLine"]] = None

    def __init__(
        self,
        id: Any,
        n: int,
        bus1: AbstractBus,
        bus2: AbstractBus,
        line_characteristics: LineCharacteristics,
        length: float,
        geometry: Optional[LineString] = None,
        **kwargs,
    ):
        """Line constructor.

        Args:
            id:
                The id of the branch.

            n:
                number of phases of the line.

            bus1:
                bus to connect to the line.

            bus2:
                bus to connect to the line.

            line_characteristics:
                The characteristics of the line.

            length:
                The length of the line in km.

            geometry:
                The geometry of the line.
        """
        if geometry is not None and not isinstance(geometry, LineString):
            msg = f"The geometry for a {type(self)} must be a linestring: {geometry.geom_type} provided."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_GEOMETRY_TYPE)

        if line_characteristics.z_line.shape != (n, n):
            msg = (
                f"Incorrect z_line dimensions for line {id!r}: {line_characteristics.z_line.shape} instead of "
                f"({n}, {n})"
            )
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_Z_LINE_SHAPE)
        if line_characteristics.y_shunt is not None and line_characteristics.y_shunt.shape != (n, n):
            msg = (
                f"Incorrect y_shunt dimensions for line {id!r}: {line_characteristics.y_shunt.shape} instead of "
                f"({n}, {n})"
            )
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_Y_SHUNT_SHAPE)

        super().__init__(n1=n, n2=n, bus1=bus1, bus2=bus2, id=id, geometry=geometry, **kwargs)
        self.n = n
        self.line_characteristics = line_characteristics

        if isinstance(length, Quantity):
            length = length.m_as("km")
        self.length = length

    def update_characteristics(self, line_characteristics: LineCharacteristics) -> None:
        """Change the line parameters.

        Args:
            line_characteristics:
                The line characteristics of the new line parameters.
        """
        self.line_characteristics = line_characteristics

    #
    # Json Mixin interface
    #
    @classmethod
    @ureg.wraps(None, (None, None, None, None, "km", None, None, None, None), strict=False)
    def from_dict(
        cls,
        id: Any,
        bus1: AbstractBus,
        bus2: AbstractBus,
        length: float,
        line_types: dict[str, LineCharacteristics],
        type_name: str,
        ground: Optional[Ground] = None,
        geometry: Optional[BaseGeometry] = None,
    ) -> "AbstractLine":
        """Line constructor from dict.

        Args:
            line_types:
                A dictionary of line characteristics by type name.

            type_name:
                The name of the line type

            id:
                The id of the created line.

            bus1:
                Bus to connect to the line.

            bus2:
                Bus to connect to the line.

            length:
                Length of the line (km).

            ground:
                The ground (optional for SimplifiedLine models).

            geometry:
                The geometry of the line.

        Returns:
            The constructed line.
        """
        line_characteristics = line_types[type_name]
        n = line_characteristics.z_line.shape[0]
        if line_characteristics.y_shunt is None:
            return cls._simplified_line_class(
                id=id,
                n=n,
                bus1=bus1,
                bus2=bus2,
                line_characteristics=line_characteristics,
                length=length,
                geometry=geometry,
            )
        else:
            return cls._shunt_line_class(
                id=id,
                n=n,
                bus1=bus1,
                bus2=bus2,
                ground=ground,
                line_characteristics=line_characteristics,
                length=length,
                geometry=geometry,
            )

    def to_dict(self) -> dict[str, Any]:
        res = super().to_dict()
        res.update(
            {
                "length": self.length,
                "type_name": self.line_characteristics.type_name,
                "type": "line",
            }
        )
        return res


class SimplifiedLine(AbstractLine):
    """A line without shunt elements.

    .. math::
        \\left(V_1 - V_2\\right) &= Z \\cdot I_1 \\\\
        I_2 &= -I_1
    """

    def __init__(
        self,
        id: Any,
        n: int,
        bus1: AbstractBus,
        bus2: AbstractBus,
        line_characteristics: LineCharacteristics,
        length: float,
        geometry: Optional[BaseGeometry] = None,
        **kwargs,
    ) -> None:
        """SimplifiedLine constructor.

        Args:
            id:
                The id of the branch.

            n:
                Number of phases of the line.

            bus1:
                Bus to connect to the line.

            bus2:
                Bus to connect to the line.

            line_characteristics:
                The characteristics of the line.

            length:
                The length of the line in km.

            geometry:
                The geometry of the line.
        """
        super().__init__(
            n=n,
            bus1=bus1,
            bus2=bus2,
            id=id,
            length=length,
            line_characteristics=line_characteristics,
            geometry=geometry,
            **kwargs,
        )

        if line_characteristics.y_shunt is not None:
            logger.warning(
                f"The simplified line {self.id!r} has been given a line characteristic "
                f"{self.line_characteristics.type_name} with a shunt. The shunt part will be ignored."
            )

    def update_characteristics(self, line_characteristics: LineCharacteristics) -> None:
        """Change the line parameters.

        Args:
            line_characteristics:
                The line characteristics of the new line parameters.
        """
        if line_characteristics.y_shunt is not None:
            logger.warning(
                f"The simplified line {self.id!r} has been given a line characteristic "
                f"{self.line_characteristics.type_name} with a shunt. The shunt part will be ignored."
            )

        super().update_characteristics(line_characteristics=line_characteristics)


class ShuntLine(AbstractLine):
    """A PI line model

    .. math::
        V_1 &= a \\cdot V_2 - b \\cdot I_2 + g \\cdot V_g \\\\
        I_1 &= c \\cdot V_2 - d \\cdot I_2 + h \\cdot V_g \\\\
        I_g &= f^t \\cdot \\left(V_1 + V_2 - 2\\cdot V_g\\right)

    with

    .. math::
        a &= \\mathcal{I}_5 + \\dfrac{1}{2} \\cdot Z \\cdot Y  \\\\
        b &= Z  \\\\
        c &= Y + \\dfrac{1}{4}\\cdot Y \\cdot Z \\cdot Y  \\\\
        d &= \\mathcal{I}_5 + \\dfrac{1}{2} \\cdot Y \\cdot Z  \\\\
        f &= -\\dfrac{1}{2} \\cdot \\begin{pmatrix} y_{ag} & y_{bg} & y_{cg} &y_{ng} \\end{pmatrix} ^t  \\\\
        g &= Z \\cdot f  \\\\
        h &= 2 \\cdot f + \\frac{1}{2}\\cdot Y \\cdot Z \\cdot f  \\\\

    """

    def __init__(
        self,
        id: Any,
        n: int,
        bus1: AbstractBus,
        bus2: AbstractBus,
        ground: Ground,
        line_characteristics: LineCharacteristics,
        length: float,
        geometry: Optional[BaseGeometry] = None,
        **kwargs,
    ) -> None:
        """ShuntLine constructor.

        Args:
            n:
                Number of phases of the line.

            bus1:
                Bus to connect to the line.

            bus2:
                Bus to connect to the line.

            ground:
                The ground.

            id:
                The id of the branch.

            line_characteristics:
                The characteristics of the line.

            length:
                The length of the line in km.

            geometry:
                The geometry of the line.
        """
        super().__init__(
            n=n,
            bus1=bus1,
            bus2=bus2,
            id=id,
            line_characteristics=line_characteristics,
            length=length,
            geometry=geometry,
            **kwargs,
        )

        self.connected_elements.append(ground)
        ground.connected_elements.append(self)


AbstractLine._simplified_line_class = SimplifiedLine
AbstractLine._shunt_line_class = ShuntLine



import numpy as np

from roseau.load_flow.utils.types import ConductorType, IsolationType, LineType
from roseau.load_flow.utils.units import Q_

PI: float = np.pi
"""The famous constant."""

MU_0: float = Q_(4 * PI * 1e-7, "H/m")
"""magnetic permeability of the vacuum (H/m)"""

EPSILON_0: float = Q_(1e-9 / (36 * PI), "F/m")
"""Relative permittivity of the vacuum (F/m)"""

F: float = Q_(50.0, "Hz")
"""Network frequency (Hz)"""

OMEGA: float = Q_(2 * PI * F, "rad/s")
"""Pulsation (rad/s)"""

RHO: dict[ConductorType, float] = {
    ConductorType.CU: Q_(1.72e-8, "ohm*m"),
    ConductorType.AL: Q_(2.82e-8, "ohm*m"),
    ConductorType.AM: Q_(3.26e-8, "ohm*m"),
    ConductorType.AA: Q_(4.0587e-8, "ohm*m"),
    ConductorType.LA: Q_(3.26e-8, "ohm*m"),
}
"""Resistivity (ohm.m)"""

CX: dict[LineType, float] = {
    LineType.OVERHEAD: Q_(0.35, "ohm/km"),
    LineType.UNDERGROUND: Q_(0.1, "ohm/km"),
    LineType.TWISTED: Q_(0.1, "ohm/km"),
}
"""Reactance parameter (Ohm/km)"""

MU_R: dict[ConductorType, float] = {
    ConductorType.CU: Q_(1.2566e-8, "H/m"),
    ConductorType.AL: Q_(1.2566e-8, "H/m"),
    ConductorType.AM: Q_(1.2566e-8, "H/m"),
    ConductorType.AA: np.nan,  # TODO
    ConductorType.LA: np.nan,
}  # TODO

"""Magnetic permeability (H/m)"""

DELTA_P: dict[ConductorType, float] = {
    ConductorType.CU: Q_(9.3, "mm"),
    ConductorType.AL: Q_(112, "mm"),
    ConductorType.AM: Q_(12.9, "mm"),
    ConductorType.AA: np.nan,  # TODO
    ConductorType.LA: np.nan,
}  # TODO

"""Skin effect (mm)"""

TAN_D: dict[IsolationType, float] = {
    IsolationType.PVC: Q_(600e-4),
    IsolationType.HDPE: Q_(6e-4),
    IsolationType.LDPE: Q_(6e-4),
    IsolationType.PEX: Q_(30e-4),
    IsolationType.EPR: Q_(125e-4),
}
"""Loss angles"""

EPSILON_R: dict[IsolationType, float] = {
    IsolationType.PVC: Q_(6.5),
    IsolationType.HDPE: Q_(2.3),
    IsolationType.LDPE: Q_(2.2),
    IsolationType.PEX: Q_(2.5),
    IsolationType.EPR: Q_(3.1),
}
"""Relative permittivity of the isolation"""

LV_MV_LIMIT: float = Q_(1000.0, "V")  # Volts
"""The limit between low-voltages and high voltages. 1000.0V exactly is low voltages."""
